#![allow(clippy::type_complexity)]

use std::ops::{Deref, DerefMut};
use std::{any::TypeId, marker::PhantomData};

use std::collections::HashMap;

use anyhow::Result;
use indexmap::IndexMap;
use update_macros::implement_updates;

use crate::resources::Resource as ResourceTrait;
use crate::*;

macro_rules! impl_update_filter {
    ( $name: ident, $([$big: ident, $little: ident]$(,)?)+ ) => {
        paste::item! {
            pub struct [<$name Filter>]<$($big: EntityComponent,)+> {
                $(
                    $little: std::marker::PhantomData<$big>,
                )+
            }

            impl<$($big: EntityComponent,)+> Default for [<$name Filter>]<$($big,)+> {
                fn default() -> Self {
                    Self {
                        $(
                            $little: std::marker::PhantomData,
                        )+
                    }
                }
            }

            impl<$($big: EntityComponent,)+> CheckFilter for [<$name Filter>]<$($big,)+> {
                fn check(entity: &EntityObject) -> bool {
                    $(
                        if entity.contains_component::<$big>() {
                            return false;
                        }
                    )+

                    true
                }

                fn verify_dedup<O: EntityComponent>() {
                    $(
                        if TypeId::of::<O>() == TypeId::of::<$big>() {
                            panic!("Type is used as input and filter at the same time");
                        }
                    )+
                }
            }

            impl<$($big: EntityComponent,)+> Clone for [<$name Filter>]<$($big,)+> {
                fn clone(&self) -> Self {
                    Self {
                        $(
                            $little: self.$little.clone(),
                        )+
                    }
                }
            }
        }
    };
}

#[derive(Copy, Clone)]
pub struct Query<C, F = EmptyFilter>
where
    F: CheckFilter,
{
    entity: Entity,
    c: C,
    filter: PhantomData<F>,
}

impl<C, F: CheckFilter> Query<C, F> {
    pub fn entity(&self) -> Entity {
        self.entity
    }
}

impl<C, F: CheckFilter> Deref for Query<C, F> {
    type Target = C;

    fn deref(&self) -> &Self::Target {
        &self.c
    }
}

impl<C, F: CheckFilter> DerefMut for Query<C, F> {
    fn deref_mut(&mut self) -> &mut Self::Target {
        &mut self.c
    }
}

pub trait AddUpdates<T, R, Func, Filter> {
    fn add_update(&mut self, name: &str, priority: u32, func: Func) -> Result<()>;
}

pub trait CreateArchetype<T, R, Func, Filter> {
    fn create(f: Func) -> Self;
}

pub trait CheckFilter: Send + Sync + Default + Clone {
    fn check(entity: &EntityObject) -> bool;
    fn verify_dedup<O: EntityComponent>();
}

#[derive(Default, Clone)]
pub struct EmptyFilter;

impl CheckFilter for EmptyFilter {
    fn check(_entity: &EntityObject) -> bool {
        true
    }

    fn verify_dedup<O>() {}
}

#[derive(Default, Clone, Debug)]
pub struct ArchetypeInfo {
    entities: Vec<(Entity, Option<String>)>,
}

impl ArchetypeInfo {
    pub fn new(entities: Vec<(Entity, Option<String>)>) -> Self {
        Self { entities }
    }

    pub fn entities(&self) -> &[(Entity, Option<String>)] {
        &self.entities
    }

    pub fn count(&self) -> usize {
        self.entities.len()
    }
}

pub struct Archetype {
    check_entity: Box<dyn Fn(&[&EntityObject]) -> bool + Send + Sync>,
    create_callback: Box<
        dyn Fn(
                &[&EntityObject],
            )
                -> Result<Box<dyn Fn(&mut Commands, &mut World) -> Result<()> + Send + Sync>>
            + Send
            + Sync,
    >,

    entities:
        IndexMap<Vec<Entity>, Box<dyn Fn(&mut Commands, &mut World) -> Result<()> + Send + Sync>>,
}

impl Archetype {
    pub fn add_entity(&mut self, entity_object: &EntityObject) -> Result<()> {
        if (self.check_entity)(entity_object) {
            let cb = (self.create_callback)(entity_object)?;

            self.entities.insert(entity_object.as_entity(), cb);
        }

        Ok(())
    }

    pub fn remove_entity(&mut self, entity: Entity) {
        self.entities.swap_remove(&entity);
    }

    pub fn execute(&self, world: &mut World) -> Result<()> {
        let mut commands = Commands::new(world.now());

        for callback in self.entities.values() {
            callback(&mut commands, world)?;
        }

        commands.apply_deferred(world)?;

        Ok(())
    }

    pub fn entities(
        &self,
    ) -> impl Iterator<
        Item = (
            &Entity,
            &Box<dyn Fn(&mut Commands, &mut World) -> Result<()> + Send + Sync>,
        ),
    > {
        self.entities.iter()
    }
}

pub struct ArchetypePair {
    check_left_entity: Box<dyn Fn(&EntityObject) -> bool + Send + Sync>,
    check_right_entity: Box<dyn Fn(&EntityObject) -> bool + Send + Sync>,

    create_callback: Box<
        dyn Fn(
                &EntityObject,
                &EntityObject,
            ) -> Result<Box<dyn Fn(&mut Commands) -> Result<()> + Send + Sync>>
            + Send
            + Sync,
    >,

    entities: IndexMap<(Entity, Entity), Box<dyn Fn(&mut Commands) -> Result<()> + Send + Sync>>,
}

impl ArchetypePair {
    pub(crate) fn add_entity(
        &mut self,
        entity_object: &EntityObject,
        entities: &IndexMap<Entity, EntityObject>,
    ) -> Result<()> {
        for (other_entity, other_entity_object) in entities.iter() {
            if entity_object.as_entity() == *other_entity {
                continue;
            }

            // check if the entities can be on both sides
            if (self.check_left_entity)(entity_object)
                && (self.check_right_entity)(other_entity_object)
            {
                let cb = (self.create_callback)(entity_object, other_entity_object)?;

                self.entities.insert(
                    (entity_object.as_entity(), other_entity_object.as_entity()),
                    cb,
                );
            }

            if (self.check_left_entity)(other_entity_object)
                && (self.check_right_entity)(entity_object)
            {
                let cb = (self.create_callback)(other_entity_object, entity_object)?;

                self.entities.insert(
                    (other_entity_object.as_entity(), entity_object.as_entity()),
                    cb,
                );
            }
        }

        Ok(())
    }

    pub(crate) fn remove_entity(&mut self, entity: Entity) {
        while let Some((left_entity, right_entity)) = self
            .entities
            .keys()
            .find(|(left_entity, right_entity)| *left_entity == entity || *right_entity == entity)
            .cloned()
        {
            self.entities.swap_remove(&(left_entity, right_entity));
        }
    }

    pub(crate) fn execute(&self, world: &mut World) -> Result<()> {
        let mut commands = Commands::new(world.now());

        for callback in self.entities.values() {
            callback(&mut commands)?;
        }

        commands.apply_deferred(world)?;

        Ok(())
    }
}

pub enum Update {
    Single(Archetype),
    Pair(ArchetypePair),
}

impl From<Archetype> for Update {
    fn from(archetype: Archetype) -> Self {
        Self::Single(archetype)
    }
}

impl From<ArchetypePair> for Update {
    fn from(pair: ArchetypePair) -> Self {
        Self::Pair(pair)
    }
}

pub struct Updates {
    updates: Vec<(u32, Update)>,
}

impl Default for Updates {
    fn default() -> Self {
        Self {
            updates: Vec::new(),
        }
    }
}

impl Updates {
    pub(crate) fn update(&mut self, world: &mut World) -> Result<()> {
        self.updates
            .iter()
            .try_for_each(|(_, update)| -> Result<()> {
                match update {
                    Update::Single(archetype) => {
                        archetype.execute(world)?;
                    }
                    Update::Pair(archetype_pair) => {
                        archetype_pair.execute(world)?;
                    }
                }

                Ok(())
            })?;

        Ok(())
    }

    pub(crate) fn add_entity(
        &mut self,
        entity_object: &EntityObject,
        entities: &IndexMap<Entity, EntityObject>,
    ) -> Result<()> {
        for (_, update) in self.updates.iter_mut() {
            match update {
                Update::Single(archetype) => {
                    archetype.add_entity(entity_object)?;
                }
                Update::Pair(archetype_pair) => {
                    archetype_pair.add_entity(entity_object, entities)?;
                }
            }
        }

        Ok(())
    }

    pub(crate) fn remove_entity(&mut self, entity: Entity) {
        for (_, update) in self.updates.iter_mut() {
            match update {
                Update::Single(archetype) => {
                    archetype.remove_entity(entity);
                }
                Update::Pair(archetype_pair) => {
                    archetype_pair.remove_entity(entity);
                }
            }
        }
    }

    // pub(crate) fn clear(&mut self) {
    //     self.updates.clear();

    //     #[cfg(feature = "timings")]
    //     self.timings.clear();
    // }

    pub(crate) fn add(&mut self, priority: u32, update: Update) -> Result<()> {
        self.updates.push((priority, update));
        self.updates
            .sort_by(|(lhs_prio, _), (rhs_prio, _)| lhs_prio.cmp(rhs_prio));

        Ok(())
    }
}

#[derive(Default)]
pub struct Archetypes {
    archetypes: HashMap<String, Archetype>,
}

impl Archetypes {
    pub(crate) fn add_entity(&mut self, entity_object: &EntityObject) -> Result<()> {
        for archetype in self.archetypes.values_mut() {
            archetype.add_entity(entity_object)?;
        }

        Ok(())
    }

    pub(crate) fn remove_entity(&mut self, entity: Entity) {
        for archetype in self.archetypes.values_mut() {
            archetype.remove_entity(entity);
        }
    }

    pub(crate) fn insert(&mut self, name: String, archetype: Archetype) {
        assert!(self.archetypes.insert(name, archetype).is_none());
    }

    pub(crate) fn get(&self, name: &String) -> Option<&Archetype> {
        self.archetypes.get(name)
    }

    pub(crate) fn execute(&self, name: &String, world: &mut World) -> Result<()> {
        self.archetypes[name].execute(world)
    }
}

implement_updates!(10, 4);

// #[rustfmt::skip]
// impl_update_filter!(Monuple,    [R, r]);
// #[rustfmt::skip]
// impl_update_filter!(Couple,     [R, r], [S, s]);
// #[rustfmt::skip]
// impl_update_filter!(Triple,     [R, r], [S, s], [T, t]);
// #[rustfmt::skip]
// impl_update_filter!(Quadruple,  [R, r], [S, s], [T, t], [U, u]);
// #[rustfmt::skip]
// impl_update_filter!(Quintuple,  [R, r], [S, s], [T, t], [U, u], [V, v]);
// #[rustfmt::skip]
// impl_update_filter!(Sextuple,   [R, r], [S, s], [T, t], [U, u], [V, v], [W, w]);
// #[rustfmt::skip]
// impl_update_filter!(Septuple,   [R, r], [S, s], [T, t], [U, u], [V, v], [W, w], [X, x]);
// #[rustfmt::skip]
// impl_update_filter!(Octuple,    [R, r], [S, s], [T, t], [U, u], [V, v], [W, w], [X, x], [Y, y]);
// #[rustfmt::skip]
// impl_update_filter!(Nonuple,    [R, r], [S, s], [T, t], [U, u], [V, v], [W, w], [X, x], [Y, y], [Z, z]);
// #[rustfmt::skip]
// impl_update_filter!(Decuple,    [R, r], [S, s], [T, t], [U, u], [V, v], [W, w], [X, x], [Y, y], [Z, z], [A, a]);
